@@ -175,7 +175,7 @@ class read_phrase ~term = object(self)
     in
     super#send_action action
 
-  method! exec = function
+  method! exec ?(keys=[]) = function
     | action :: actions when S.value self#mode = LTerm_read_line.Edition &&
                              is_accept action  -> begin
         Zed_macro.add self#macro action;
@@ -205,11 +205,11 @@ class read_phrase ~term = object(self)
               UTop_history.add_warnings UTop.stashable_session_history warnings;
               UTop_history.add_error UTop.stashable_session_history msg;
           end;
-          return result
+          return (LTerm_read_line.Result result)
         with UTop.Need_more ->
           (* Input not finished, continue. *)
           self#insert (UChar.of_char '\n');
-          self#exec actions
+          self#exec ~keys actions
       end
     | actions ->
       super_term#exec actions
@@ -437,9 +437,17 @@ type rewrite_rule = {
   (* Whether the rule is enabled or not. *)
 }
 
+#if OCAML_VERSION < (4, 11, 0)
+let longident_parse= Longident.parse
+#else
+let longident_parse str=
+  let lexbuf= Lexing.from_string str in
+  Parse.longident lexbuf
+#endif
+
 let longident_lwt_main_run = Longident.Ldot (Longident.Lident "Lwt_main", "run")
 let longident_async_thread_safe_block_on_async_exn =
-  Longident.parse "Async.Thread_safe.block_on_async_exn"
+  Longident.(Ldot (Ldot (Lident "Async", "Thread_safe"), "block_on_async_exn"))
 let longident_unit = Longident.Lident "()"
 
 #if OCAML_VERSION >= (4, 03, 0)
@@ -451,7 +459,7 @@ let nolabel = ""
 let rewrite_rules = [
   (* Rewrite Lwt.t expressions to Lwt_main.run <expr> *)
   {
-    type_to_rewrite = Longident.parse "Lwt.t";
+    type_to_rewrite = Longident.(Ldot (Lident "Lwt", "t"));
     path_to_rewrite = None;
     required_values = [longident_lwt_main_run];
     rewrite = (fun loc e ->
@@ -466,7 +474,7 @@ let rewrite_rules = [
   (* Rewrite Async.Defered.t expressions to
      Async.Thread_safe.block_on_async_exn (fun () -> <expr>). *)
   {
-    type_to_rewrite = Longident.parse "Async.Deferred.t";
+    type_to_rewrite = Longident.(Ldot (Ldot (Lident "Async", "Deferred"), "t"));
     path_to_rewrite = None;
     required_values = [longident_async_thread_safe_block_on_async_exn];
     rewrite = (fun loc e ->
@@ -710,9 +718,20 @@ let execute_phrase = Toploop.execute_phrase
    | Main loop                                                       |
    +-----------------------------------------------------------------+ *)
 
+let registers= ref LTerm_vi.Vi.Interpret.RegisterMap.empty
+
 let rec read_phrase term =
   Lwt.catch
-    (fun () -> (new read_phrase ~term)#run)
+    (fun () ->
+      let read_line= new read_phrase ~term in
+      (match !UTop.edit_mode with
+      | LTerm_editor.Default-> ()
+      | LTerm_editor.Vi as mode-> read_line#set_editor_mode mode);
+      let vi_state= read_line#vi_state in
+      vi_state#set_registers !registers;
+      read_line#run >>= fun result->
+      registers:= vi_state#get_registers;
+      return result)
     (function
     | Sys.Break ->
       LTerm.fprintl term "Interrupted." >>= fun () ->
@@ -1177,7 +1196,7 @@ end
    +-----------------------------------------------------------------+ *)
 
 let typeof sid =
-  let id  = Longident.parse sid in
+  let id  = longident_parse sid in
   let env = !Toploop.toplevel_env in
 #if OCAML_VERSION >= (4, 10, 0)
   let lookup_value= Env.find_value_by_name
@@ -1449,20 +1468,34 @@ let common_init ~initial_env =
       if !autoload && !UTop_private.autoload && Sys.file_exists autoload_dir then
         load_init_files autoload_dir
     | None -> ());
-  (* Load user's .ocamlinit file. *)
-  (match !Clflags.init_file with
+  (* Load user's init file. *)
+  let init_fn =
+    match !Clflags.init_file with
      | Some fn ->
          if Sys.file_exists fn then
-           ignore (Toploop.use_silently Format.err_formatter fn : bool)
-         else
-           Printf.eprintf "Init file not found: \"%s\".\n" fn
+           Some fn
+         else (
+           Printf.eprintf "Init file not found: \"%s\".\n" fn;
+           None
+         )
      | None ->
          if Sys.file_exists ".ocamlinit" then
-           ignore (Toploop.use_silently Format.err_formatter ".ocamlinit" : bool)
+           Some ".ocamlinit"
          else
-           let fn = Filename.concat LTerm_resources.home ".ocamlinit" in
-           if Sys.file_exists fn then
-             ignore (Toploop.use_silently Format.err_formatter fn));
+           let xdg_fn = LTerm_resources.xdgbd_file ~loc:LTerm_resources.Config "utop/init.ml" in
+           if Sys.file_exists xdg_fn then
+             Some xdg_fn
+           else
+             let fn = Filename.concat LTerm_resources.home ".ocamlinit" in
+             if Sys.file_exists fn then
+               Some fn
+             else
+               None
+  in
+  (match init_fn with
+   | None -> ()
+   | Some fn ->
+     ignore (Toploop.use_silently Format.err_formatter fn : bool));
   (* Load history after the initialization file so the user can change
      the history file name. *)
   Lwt_main.run (init_history ());
